var typeCheck, equal;
typeCheck = require('../..');
equal = require('assert').strictEqual;
describe('index', function(){
  it('version', function(){
    equal(typeCheck.VERSION, require('../../package.json').version);
  });
});
